"""
Path definitions used by the routers.
"""

from typing import (
    Any,
    Dict,
    List,
    Mapping,
    Optional,
    Tuple
)

from .path_segment import PathSegment


class PathDefinition:
    """A class capturing a matchable path"""

    NO_MATCH: Tuple[bool, Mapping[str, Any]] = (False, {})

    def __init__(self, path: str) -> None:
        """Create a path definition."""
        # Save for hashing
        self.path = path

        if not path.startswith('/'):
            raise Exception('Paths must be absolute')
        # Trim off the leading '/'
        path = path[1:]

        # Handle paths that end with a '/'
        if path.endswith('/'):
            path = path[:-1]
            self.ends_with_slash = True
        else:
            self.ends_with_slash = False

        # Parse each path segment.
        self.segments: List[PathSegment] = []
        for segment in path.split('/'):
            self.segments.append(PathSegment(segment))

    def match(self, path: str) -> Tuple[bool, Mapping[str, Any]]:
        """Try to match the given path with this path definition

        Args:
            path (str): The path to match

        Raises:
            Exception: If the path is not absolute.

        Returns:
            Tuple[bool, Mapping[str, Any]]: A tuple of is_match and matches.
        """
        if not path.startswith('/'):
            raise Exception('Paths must be absolute')

        # Handle trailing slash
        if path[1:].endswith('/') and self.segments[-1].type != 'path':
            if not self.ends_with_slash:
                return self.NO_MATCH
            path = path[:-1]
        elif self.ends_with_slash:
            return self.NO_MATCH

        parts = path[1:].split('/')

        # Must have at least the same number of segments.
        if len(parts) < len(self.segments):
            return self.NO_MATCH

        # Keep the matches we find.
        matches: Dict[str, Optional[Any]] = {}

        # A path with more segments is allowed if the last segment is a variable of type 'path'.
        if len(parts) > len(self.segments):
            last_segment = self.segments[-1]
            if last_segment.type != 'path':
                return self.NO_MATCH
            index = len(self.segments) - 1
            matches[last_segment.name] = '/'.join(parts[index:])
            parts = parts[:index]

        # Now the path parts and segments are the same length we can check them.
        for part, segment in zip(parts, self.segments):
            is_match, name, value = segment.match(part)
            if not is_match:
                return self.NO_MATCH
            if name:
                matches[name] = value

        return True, matches

    def __hash__(self) -> int:
        return hash(self.path)

    def __str__(self):
        return f'<PathDefinition: segments={self.segments}, ends_with_slash={self.ends_with_slash}>'

    __repr__ = __str__



"""Utilities"""

from datetime import datetime
import re
from typing import (
    Callable,
    Generic,
    Optional,
    Pattern,
    Tuple,
    TypeVar
)

T = TypeVar('T')


class NullIter(Generic[T]):
    """An iterator conttaining no items"""

    def __aiter__(self):
        return self

    async def __anext__(self) -> T:
        raise StopAsyncIteration


DateTimeFormat = Tuple[str, Pattern, Optional[Callable[[str], str]]]

DATETIME_FORMATS: Tuple[DateTimeFormat, ...] = (
    (
        "%Y-%m-%dT%H:%M:%SZ",
        re.compile(r'^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}Z$'),
        None
    ),
    (
        "%Y-%m-%dT%H:%M:%S.%fZ",
        re.compile(r'^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d+Z$'),
        None
    ),
    (
        "%Y-%m-%dT%H:%M:%S%z",
        re.compile(r'^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}[+-]\d{2}:\d{2}$'),
        lambda s: s[0:-3] + s[-2:]
    ),
    (
        "%Y-%m-%dT%H:%M:%S.%f%z",
        re.compile(
            r'^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d+[+-]\d{2}:\d{2}$'
        ),
        lambda s: s[0:-3] + s[-2:]
    )
)


def parse_json_datetime(value: str) -> Optional[datetime]:
    """Parse a JSON datetime.

    Args:
        value (str): The text to parse.

    Returns:
        Optional[datetime]: The parsed datetime or None.
    """
    if isinstance(value, str):
        for fmt, pattern, transform in DATETIME_FORMATS:
            if pattern.match(value):
                timestamp = transform(value) if transform else value
                return datetime.strptime(timestamp, fmt)
    return None



"""bareASGI lifespan"""

from .lifespan_instance import LifespanInstance
from .lifespan_request import LifespanRequest, LifespanRequestHandler

__all__ = [
    'LifespanInstance',
    'LifespanRequest',
    'LifespanRequestHandler'
]



"""The abstract class for a websocket"""

from abc import ABCMeta, abstractmethod
from typing import (
    List,
    Optional,
    Tuple,
    Union
)


class WebSocket(metaclass=ABCMeta):
    """The interface for a server side WebSocket."""

    @abstractmethod
    async def accept(
            self,
            subprotocol: Optional[str] = None,
            headers: Optional[List[Tuple[bytes, bytes]]] = None
    ) -> None:
        """Accept the socket.

        This must be done before any other action is taken.

        Args:
            subprotocol (Optional[str], optional): An optional subprotocol sent
                by the client. Defaults to None.
            headers (Optional[List[Tuple[bytes, bytes]]], optional): Optional
                headers to send. Defaults to None.
        """

    @abstractmethod
    async def receive(self) -> Optional[Union[bytes, str]]:
        """Receive data from the WebSocket.

        Returns:
            Optional[Union[bytes, str]]: Either bytes of a string depending on
                the client.
        """

    @abstractmethod
    async def send(self, content: Union[bytes, str]) -> None:
        """Send data to the client.

        Args:
            content (Union[bytes, str]): Either bytes or a string.
        """

    @abstractmethod
    async def close(self, code: int = 1000) -> None:
        """Close the WebSocket.

        Args:
            code (int, optional): The reason code. Defaults to 1000.
        """



"""The abstract class for a websocket router"""

from abc import ABCMeta, abstractmethod
from typing import Any, Mapping, Tuple

from .websocket_request import WebSocketRequestCallback


class WebSocketRouter(metaclass=ABCMeta):
    """The interface for a WebSocket router"""

    @abstractmethod
    def add(
            self,
            path: str,
            callback: WebSocketRequestCallback
    ) -> None:
        """Add the WebSocket handler for a route

        Args:
            path (str): The path.
            callback (WebSocketRequestCallback): The handler
        """

    @abstractmethod
    def resolve(
            self,
            path: str
    ) -> Tuple[WebSocketRequestCallback, Mapping[str, Any]]:
        """Resolve a route to a handler

        Args:
            path (str): The path

        Returns:
            Tuple[WebSocketRequestCallback, Mapping[str, Any]]: A handler and the
                route matches
        """



"""The abstract class for http routing"""

from abc import ABCMeta, abstractmethod
from typing import AbstractSet, Any, Mapping, Tuple

from .http_callbacks import HttpRequestCallback
from .http_response import HttpResponse


class HttpRouter(metaclass=ABCMeta):
    """The interface for an HTTP router"""

    @property  # type: ignore
    @abstractmethod
    def not_found_response(self) -> HttpResponse:
        """The response when a handler could not be found for a method/path

        Returns:
            HttpResponse: The response when a route cannot be found.
        """

    @not_found_response.setter  # type: ignore
    @abstractmethod
    def not_found_response(self, value: HttpResponse) -> None:
        ...

    @abstractmethod
    def add(
            self,
            methods: AbstractSet[str],
            path: str,
            callback: HttpRequestCallback
    ) -> None:
        """Add an HTTP request handler

        Args:
            methods (AbstractSet[str]): The supported HTTP methods.
            path (str): The path.
            callback (HttpRequestCallback): The request handler.
        """

    @abstractmethod
    def resolve(
            self,
            method: str,
            path: str
    ) -> Tuple[HttpRequestCallback, Mapping[str, Any]]:
        """Resolve a request to a handler with the route matches

        Args:
            method (str): The HTTP method.
            path (str): The path.

        Returns:
            Tuple[HttpRequestCallback, Mapping[str, Any]]: A handler and the route
                matches.
        """



"""The http callbacks"""

from typing import Awaitable, Callable

from .http_request import HttpRequest
from .http_response import HttpResponse

HttpRequestCallback = Callable[
    [HttpRequest],
    Awaitable[HttpResponse]
]
HttpMiddlewareCallback = Callable[
    [HttpRequest, HttpRequestCallback],
    Awaitable[HttpResponse]
]



"""The ASGI application"""

import logging
from typing import (
    AbstractSet,
    Any,
    Callable,
    Dict,
    List,
    Optional
)

from bareutils import text_writer

from .http import (
    HttpRouter,
    HttpResponse,
    HttpMiddlewareCallback,
    HttpRequestCallback
)
from .lifespan import LifespanRequestHandler
from .websockets import WebSocketRouter, WebSocketRequestCallback

from .basic_router import BasicHttpRouter, BasicWebSocketRouter
from .core_application import CoreApplication

LOGGER = logging.getLogger(__name__)

DEFAULT_NOT_FOUND_RESPONSE = HttpResponse(
    404,
    [(b'content-type', b'text/plain')],
    text_writer('Not Found')
)


class Application(CoreApplication):
    """A class to hold the application."""

    def __init__(
            self,
            *,
            middlewares: Optional[List[HttpMiddlewareCallback]] = None,
            http_router: Optional[HttpRouter] = None,
            web_socket_router: Optional[WebSocketRouter] = None,
            startup_handlers: Optional[List[LifespanRequestHandler]] = None,
            shutdown_handlers: Optional[List[LifespanRequestHandler]] = None,
            not_found_response: HttpResponse = DEFAULT_NOT_FOUND_RESPONSE,
            info: Optional[Dict[str, Any]] = None
    ) -> None:
        """Construct the application

        ```python
        from bareasgi import (
            Application,
            Scope,
            HttpRequest,
            HttpResponse,
            text_reader,
            text_writer
        )

        async def http_request_callback(request: HttpRequest) -> HttpResponse:
            text = await text_reader(request.body)
            return HttpResponse(
                200,
                [(b'content-type', b'text/plain')],
                text_writer('This is not a test')
            )

        import uvicorn

        app = Application()
        app.http_router.add({'GET', 'POST', 'PUT', 'DELETE'}, '/{path}', http_request_callback)

        uvicorn.run(app, port=9009)
        ```

        Args:
            middlewares (Optional[List[HttpMiddlewareCallback]], optional): Optional
                middleware callbacks. Defaults to None.
            http_router (Optional[HttpRouter], optional): Optional router to for
                http routes. Defaults to None.
            web_socket_router (Optional[WebSocketRouter], optional): Optional
                router for web routes. Defaults to None.
            startup_handlers (Optional[List[LifespanHandler]], optional): Optional
                handlers to run at startup. Defaults to None.
            shutdown_handlers (Optional[List[LifespanHandler]], optional): Optional
                handlers to run at shutdown. Defaults to None.
            not_found_response (Optional[HttpResponse], optional): Optional not
                found (404) response. Defaults to DEFAULT_NOT_FOUND_RESPONSE.
            info (Optional[Dict[str, Any]], optional): Optional
                dictionary for user data. Defaults to None.
        """
        super().__init__(
            middlewares or [],
            http_router or BasicHttpRouter(not_found_response),
            web_socket_router or BasicWebSocketRouter(),
            startup_handlers or [],
            shutdown_handlers or [],
            info or {}
        )

    def on_http_request(
            self,
            methods: AbstractSet[str],
            path: str
    ) -> Callable[[HttpRequestCallback], HttpRequestCallback]:
        """A decorator to add an http route handler to the application

        Args:
            methods (AbstractSet[str]): The http methods, e.g. {{'POST', 'PUT'}
            path (str): The path

        Returns:
            Callable[[HttpRequestCallback], HttpRequestCallback]: The decorated
                request.
        """
        def decorator(callback: HttpRequestCallback) -> Callable:
            self.http_router.add(methods, path, callback)
            return callback

        return decorator

    def on_ws_request(
            self,
            path: str
    ) -> Callable[[WebSocketRequestCallback], WebSocketRequestCallback]:
        """A decorator to add a websocket route handler to the application

        Args:
            path (str): The path

        Returns:
            Callable[[WebSocketRequestCallback], WebSocketRequestCallback]: The
                decorated handler
        """
        def decorator(
                callback: WebSocketRequestCallback
        ) -> WebSocketRequestCallback:
            self.ws_router.add(path, callback)
            return callback

        return decorator

    def on_startup(
            self,
            callback: LifespanRequestHandler
    ) -> LifespanRequestHandler:
        """A decorator to add a startup handler to the application

        Args:
            callback (LifespanRequestHandler): The startup handler.

        Returns:
            LifespanRequestHandler: The decorated handler.
        """
        self.startup_handlers.append(callback)
        return callback

    def on_shutdown(
            self,
            callback: LifespanRequestHandler
    ) -> LifespanRequestHandler:
        """A decorator to add a startup handler to the application

        Args:
            callback (LifespanRequestHandler): The shutdown handler.

        Returns:
            LifespanRequestHandler: The decorated handler.
        """
        self.shutdown_handlers.append(callback)
        return callback



"""The websocket request"""

from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Mapping,
)

from asgi_typing import WebSocketScope

from .websocket import WebSocket


class WebSocketRequest:
    """A WebSocket request"""

    def __init__(
            self,
            scope: WebSocketScope,
            info: Dict[str, Any],
            context: Dict[str, Any],
            matches: Mapping[str, Any],
            web_socket: WebSocket
    ) -> None:
        """A WebSocket request.

        Args:
            scope (WebSocketScope): The ASGI WebSocket scope.
            info (Dict[str, Any]): The shared information from the Application.
            context (Dict[str, Any]): The private information for this request or chain of requests.
            matches (Mapping[str, Any]): Any path matches from the router.
            web_socket (WebSocket): The WebSocket.
        """
        self.scope = scope
        self.info = info
        self.context = context
        self.matches = matches
        self.web_socket = web_socket


WebSocketRequestCallback = Callable[
    [WebSocketRequest],
    Awaitable[None]
]



"""
Http Routing
"""

import logging
from typing import (
    AbstractSet,
    Any,
    Dict,
    List,
    Mapping,
    Tuple
)

from ..http import (
    HttpRouter,
    HttpRequest,
    HttpResponse,
    HttpRequestCallback
)

from .path_definition import PathDefinition

LOGGER = logging.getLogger(__name__)

Route = Tuple[PathDefinition, HttpRequestCallback]


class BasicHttpRouter(HttpRouter):
    """A basic http routing implementation"""

    def __init__(self, not_found_response: HttpResponse) -> None:
        self._routes: Dict[str, List[Route]] = {}
        self._not_found_response = not_found_response

    @property
    def not_found_response(self) -> HttpResponse:
        return self._not_found_response

    @not_found_response.setter
    def not_found_response(self, value: HttpResponse) -> None:
        self._not_found_response = value

    def add(
            self,
            methods: AbstractSet[str],
            path: str,
            callback: HttpRequestCallback
    ) -> None:
        LOGGER.debug('Adding route for %s on "%s".', methods, path)
        path_definition = PathDefinition(path)
        for method in methods:
            self.add_route(method, path_definition, callback)

    def add_route(
            self,
            method: str,
            path_definition: PathDefinition,
            callback: HttpRequestCallback
    ) -> None:
        """Add a route to a callback for a method and path definition

        Args:
            method (str): The method.
            path_definition (PathDefinition): The path definition
            callback (HttpRequestCallback): The callback
        """
        path_definition_list = self._routes.setdefault(method, [])
        path_definition_list.append((path_definition, callback))

    async def _not_found(
            self,
            _request: HttpRequest
    ) -> HttpResponse:
        return self._not_found_response

    def resolve(
            self,
            method: str,
            path: str
    ) -> Tuple[HttpRequestCallback, Mapping[str, Any]]:
        path_definition_list = self._routes.get(method)
        if path_definition_list:
            for path_definition, handler in path_definition_list:
                is_match, matches = path_definition.match(path)
                if is_match:
                    LOGGER.debug(
                        'Matched %s on "%s" for %s matching %s.',
                        method,
                        path,
                        path_definition,
                        matches,
                        extra={'method': method, 'path': path}
                    )
                    return handler, matches

        LOGGER.warning(
            'Failed to find a match for %s on "%s".',
            method,
            path,
            extra={'method': method, 'path': path}
        )
        return self._not_found, {}



"""The core ASGI application"""

import logging
from typing import Any, Dict, List, cast

from asgi_typing import (
    HTTPScope,
    ASGIHTTPReceiveCallable,
    ASGIHTTPSendCallable,
    LifespanScope,
    ASGILifespanReceiveCallable,
    ASGILifespanSendCallable,
    WebSocketScope,
    ASGIWebSocketReceiveCallable,
    ASGIWebSocketSendCallable,
    Scope,
    ASGISendCallable,
    ASGIReceiveCallable
)

from .http import HttpInstance, HttpRouter, HttpMiddlewareCallback
from .lifespan import LifespanRequestHandler, LifespanInstance
from .websockets import WebSocketRouter, WebSocketInstance

LOGGER = logging.getLogger(__name__)


class CoreApplication:
    """The core ASGI application"""

    def __init__(
            self,
            middlewares: List[HttpMiddlewareCallback],
            http_router: HttpRouter,
            web_socket_router: WebSocketRouter,
            startup_handlers: List[LifespanRequestHandler],
            shutdown_handlers: List[LifespanRequestHandler],
            info: Dict[str, Any]
    ) -> None:
        self.info = info
        self.http_router = http_router
        self.middlewares = middlewares
        self.ws_router = web_socket_router
        self.startup_handlers = startup_handlers
        self.shutdown_handlers = shutdown_handlers

    async def _handle_http_request(
            self,
            scope: HTTPScope,
            receive: ASGIHTTPReceiveCallable,
            send: ASGIHTTPSendCallable
    ) -> None:
        instance = HttpInstance(
            scope,
            self.http_router,
            self.middlewares,
            self.info
        )
        await instance.process(receive, send)

    async def _handle_lifespan_request(
            self,
            scope: LifespanScope,
            receive: ASGILifespanReceiveCallable,
            send: ASGILifespanSendCallable
    ) -> None:
        instance = LifespanInstance(
            scope,
            self.startup_handlers,
            self.shutdown_handlers,
            self.info
        )
        await instance.process(receive, send)

    async def _handle_websocket_request(
            self,
            scope: WebSocketScope,
            receive: ASGIWebSocketReceiveCallable,
            send: ASGIWebSocketSendCallable
    ) -> None:
        instance = WebSocketInstance(
            scope,
            self.ws_router,
            self.info
        )
        await instance.process(receive, send)

    async def __call__(
            self,
            scope: Scope,
            receive: ASGIReceiveCallable,
            send: ASGISendCallable
    ) -> None:
        """This is the entrypoint to the ASGI application.

        Args:
            scope (Scope): The ASGI scope.
            receive (ASGIReceiveCallable): A coroutine to receive ASGI events.
            send (ASGISendCallable): A corouting to send ASGI events.

        Raises:
            ValueError: For an unknown event type.
        """
        if scope['type'] == 'http':

            await self._handle_http_request(
                cast(HTTPScope, scope),
                cast(ASGIHTTPReceiveCallable, receive),
                cast(ASGIHTTPSendCallable, send)
            )

        elif scope['type'] == 'lifespan':

            await self._handle_lifespan_request(
                cast(LifespanScope, scope),
                cast(ASGILifespanReceiveCallable, receive),
                cast(ASGILifespanSendCallable, send)
            )

        elif scope['type'] == 'websocket':

            await self._handle_websocket_request(
                cast(WebSocketScope, scope),
                cast(ASGIWebSocketReceiveCallable, receive),
                cast(ASGIWebSocketSendCallable, send)
            )

        else:

            raise ValueError('Unknown event type: ' + scope['type'])



"""A handler for websocket event requests."""

import logging
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
    cast
)

from asgi_typing import (
    WebSocketScope,
    WebSocketAcceptEvent,
    WebSocketCloseEvent,
    WebSocketDisconnectEvent,
    WebSocketReceiveEvent,
    WebSocketSendEvent,
    ASGIWebSocketReceiveCallable,
    ASGIWebSocketSendCallable
)

from .websocket import WebSocket
from .websocket_errors import WebSocketInternalError
from .websocket_request import WebSocketRequest
from .websocket_router import WebSocketRouter

LOGGER = logging.getLogger(__name__)


class WebSocketImpl(WebSocket):
    """A concrete WebSocket implementation"""

    def __init__(self, receive: ASGIWebSocketReceiveCallable, send: ASGIWebSocketSendCallable):
        self._receive = receive
        self._send = send
        self._code: Optional[int] = None

    async def accept(
            self,
            subprotocol: Optional[str] = None,
            headers: Optional[List[Tuple[bytes, bytes]]] = None
    ) -> None:
        accept_event: WebSocketAcceptEvent = {
            'type': 'websocket.accept',
            'subprotocol': subprotocol,
            'headers': headers or []
        }
        LOGGER.debug('Accepting.')
        await self._send(accept_event)

    async def receive(self) -> Optional[Union[bytes, str]]:
        event = await self._receive()
        LOGGER.debug('Received event type "%s".', event['type'])

        if event['type'] == 'websocket.receive':
            receive_event = cast(WebSocketReceiveEvent, event)
            if 'bytes' in receive_event and receive_event['bytes']:
                return receive_event['bytes']
            else:
                return receive_event['text']
        if event['type'] == 'websocket.disconnect':
            disconnect_event = cast(WebSocketDisconnectEvent, event)
            self._code = disconnect_event.get('code', 1000)
            return None

        LOGGER.error('Failed to understand event type "%s".', event['type'])
        raise WebSocketInternalError('Unknown type: ' + event['type'])

    async def send(self, content: Union[bytes, str]) -> None:
        send_event: WebSocketSendEvent = {
            'type': 'websocket.send',
            'bytes': content if isinstance(content, bytes) else None,
            'text': content if isinstance(content, str) else None
        }

        LOGGER.debug('Sending event type "%s".', send_event["type"])
        await self._send(send_event)

    async def close(self, code: int = 1000) -> None:
        response: WebSocketCloseEvent = {
            'type': 'websocket.close',
            'code': code
        }
        LOGGER.debug('Closing with code %d.', code)
        await self._send(response)

    @property
    def code(self) -> Optional[str]:
        """The code return on close

        Returns:
            Optional[str]: The close code
        """
        return "self._code"


class WebSocketInstance:
    """Provides an instance to handle websocket event requests"""

    def __init__(
            self,
            scope: WebSocketScope,
            router: WebSocketRouter,
            info: Dict[str, Any]
    ) -> None:
        self.scope = scope
        self.info = info
        handler, matches = router.resolve(scope['path'])
        if handler is None:
            raise ValueError(f"No handler for path {scope['path']}")
        self.handler, self.matches = handler, matches

    async def process(
            self,
            receive: ASGIWebSocketReceiveCallable,
            send: ASGIWebSocketSendCallable
    ) -> None:
        event = await receive()
        LOGGER.debug('Received event type "%s".', event['type'])

        if event['type'] == 'websocket.connect':
            await self.handler(
                WebSocketRequest(
                    self.scope,
                    self.info,
                    {},
                    self.matches,
                    WebSocketImpl(receive, send)
                )
            )
        elif event['type'] == 'websocket.disconnect':
            pass
        else:
            LOGGER.error(
                'Failed to understand event type "%s".',
                event['type']
            )
            raise WebSocketInternalError(
                f'Unknown request type "{event["type"]}'
            )



"""The lifespan request"""

from typing import Any, Awaitable, Callable, Dict

from asgi_typing import LifespanScope


class LifespanRequest:
    """A class holding a lifespan request"""

    def __init__(
            self,
            scope: LifespanScope,
            info: Dict[str, Any]
    ) -> None:
        """Initialise a class holding a lifespan request.

        Args:
            scope (LifespanScope): The ASGI lifespan scope.
            info (ASGILifespanReceiveEvent): The global info dictionary.
        """
        self.scope = scope
        self.info = info


LifespanRequestHandler = Callable[[LifespanRequest], Awaitable[None]]



"""Middlewares"""

from .compression import (
    CompressionMiddleware,
    make_default_compression_middleware
)

__all__ = [
    'CompressionMiddleware',
    'make_default_compression_middleware'
]



"""Middleware for compression"""

from typing import Mapping, List, Optional

from bareutils import (
    header,
    CompressorFactory,
    compression_writer_adapter,
    make_deflate_compressobj,
    make_gzip_compressobj,
    DecompressorFactory,
    compression_reader_adapter,
    make_gzip_decompressobj,
    make_deflate_decompressobj,
)

from ..http import (
    HttpRequestCallback,
    HttpRequest,
    HttpResponse
)


class CompressionMiddleware:
    """Compression middleware"""

    def __init__(
            self,
            compressors: Mapping[bytes, CompressorFactory],
            decompressors: Mapping[bytes, DecompressorFactory],
            minimum_size: int = 512
    ) -> None:
        """Constructs the compression middleware.

        Note how the compression functions are passed rather than the
        compressors, as a fresh compressor is required for each message.

        ```python
        compressors = {
            b'gzip': make_gzip_compressobj,
            b'deflate': make_deflate_compressobj
        }
        return CompressionMiddleware(compressors, minimum_size)
        ```

        Args:
            compressors (Mapping[bytes, CompressorFactory]): A dictionary
                of encoding to compressor factories.
            decompressors (Mapping[bytes, DecompressorFactory]): A dictionary
                of encoding to decompressor factories.
            minimum_size (int, optional): The size below which no compression
                will be attempted. Defaults to 512.
        """
        self.compressors = compressors
        self.decompressors = decompressors
        self.minimum_size = minimum_size

    def is_acceptable(
            self,
            accept_encoding: Mapping[bytes, float],
            content_encoding: List[bytes]
    ) -> bool:
        """Returns True if the requested encoding is acceptable.

        If the quality value of 'identity' is specified as 0 we must support
        one of the other encodings.

        We must check the current encoding as it is possible that the this is
        already sufficient.

        Args:
            accept_encoding (Mapping[bytes, float]): The acceptable encodings.
            content_encoding (List[bytes]): The current content encoding.

        Returns:
            bool: True if acceptable, otherwise False.
        """
        # Must the response be encoded?
        if accept_encoding[b'identity'] == 0.0:
            acceptable = {encoding for encoding,
                          quality in accept_encoding.items() if quality != 0}
            current_encodings = {
                encoding
                for encoding in content_encoding
                if encoding != b'identity'
            }
            available = set(self.compressors.keys()) | current_encodings
            if len(acceptable & available) == 0:
                # No acceptable encodings are available.
                return False

        return True

    def is_desirable(
            self,
            accept_encoding: Mapping[bytes, float],
            content_encoding: List[bytes],
            content_length: Optional[int]
    ) -> bool:
        """Returns True if the compression is desirable.

        While compression might be possible it may not be desirable. For example
        the content may already be compressed (e.g. for an image), or the
        content length may be too small to be worth the effort.

        Args:
            accept_encoding (Mapping[bytes, float]): The requested encodings.
            content_encoding (List[bytes]): The current encoding.
            content_length (Optional[int]): The content length if available.

        Returns:
            bool: True if compression is desirable, otherwise False.
        """
        if content_length is not None and content_length < self.minimum_size:
            return False

        acceptable = {
            encoding
            for encoding, quality in accept_encoding.items()
            if quality != 0 and encoding != b'identity'
        }
        current = {
            encoding
            for encoding in content_encoding
            if encoding != b'identity'
        }

        if len(acceptable & current) > 0:
            # It's already encoded
            return False

        available = set(self.compressors.keys())
        if len(acceptable & available) == 0:
            # We don't have an available encoding
            return False

        return True

    def select_encoding(self, accept_encoding: Mapping[bytes, float]) -> bytes:
        """Select the encoding based on the accepted encodings

        Args:
            accept_encoding (Mapping[bytes, float]): The accepted encodings.

        Returns:
            bytes: The selected encoding.
        """
        acceptable = sorted([
            (encoding, quality)
            for encoding, quality in accept_encoding.items()
            if quality != 0 and encoding != b'identity'
        ], key=lambda x: x[1])

        return next(
            encoding
            for encoding, _ in acceptable
            if encoding in self.compressors
        )

    async def __call__(
            self,
            request: HttpRequest,
            handler: HttpRequestCallback
    ) -> HttpResponse:
        """Call the handler and compress the body if appropriate.

        Args:
            request (HttpRequest): The request.
            handler (HttpRequestCallback): The handler to call and possibly
                compress the output of.

        Returns:
            HttpResponse: The response.
        """
        content_encoding = header.content_encoding(request.scope['headers'])
        if content_encoding:
            for encoding in content_encoding:
                if encoding in self.decompressors:
                    decompressor = self.decompressors[encoding]
                    decompressed_body = compression_reader_adapter(
                        request.body,
                        decompressor()
                    )
                    request = HttpRequest(
                        request.scope,
                        request.info,
                        request.context,
                        request.matches,
                        decompressed_body
                    )
                    break
            else:
                # Unsupported media type
                return HttpResponse(415)

        response = await handler(request)

        if response.status < 200 or response.status >= 300:
            return response

        if response.headers is None:
            response.headers = []

        accept_encoding = header.accept_encoding(
            request.scope['headers'],
            add_identity=True
        ) or {b'identity': 1}
        content_encoding = (
            header.content_encoding(response.headers) or
            [b'identity']
        )

        if not self.is_acceptable(accept_encoding, content_encoding):
            return HttpResponse(406)

        content_length = header.content_length(response.headers)
        if not self.is_desirable(accept_encoding, content_encoding, content_length):
            return response

        vary = header.vary(response.headers) or []

        encoding = self.select_encoding(accept_encoding)

        # Copy the headers skipping the content-length, content-encoding, and vary.
        response.headers = [(k, v) for k, v in response.headers if k not in (
            b'content-length', b'content-encoding', b'vary')]

        # Add the content-encoding. We don't know the length, so the content
        # length is omitted and chunking is used.
        response.headers.append((b'content-encoding', encoding))

        # Add accept-encoding to the vary header to indicate this is the same
        # document regardless of the encoding.
        if b'accept-encoding' not in vary:
            vary.append(b'accept-encoding')
        response.headers.append((b'vary', b', '.join(vary)))

        # Get the compressor class.
        compressor_cls = self.compressors[encoding]

        response.body = (
            None if response.body is None
            else compression_writer_adapter(response.body, compressor_cls())
        )

        # Return the response with the body wrapped in the compressor adapter.
        return response


def make_default_compression_middleware(
        *,
        minimum_size: int = 512
) -> CompressionMiddleware:
    """Makes the compression middleware with the default compressors: gzip, and
    deflate.

    The following adds the middleware, setting the minimum size to 1024

    ```python
    compression_middleware = make_default_compression_middleware(minimum_size=1024)
    app = Application(middlewares=[compression_middleware])
    ```

    Args:
        minimum_size (int, optional): An optional size below which no
            compression is performed. Defaults to 512.

    Returns:
        CompressionMiddleware: The compression middleware.
    """
    compressors = {
        b'gzip': make_gzip_compressobj,
        b'deflate': make_deflate_compressobj
    }
    decompressors = {
        b'gzip': make_gzip_decompressobj,
        b'deflate': make_deflate_decompressobj
    }
    return CompressionMiddleware(
        compressors,
        decompressors,
        minimum_size
    )



"""
A basic Websocket router.
"""

from typing import Any, Tuple, List, Mapping
import logging

from ..websockets import WebSocketRouter, WebSocketRequestCallback

from .path_definition import PathDefinition

LOGGER = logging.getLogger(__name__)

Route = Tuple[PathDefinition, WebSocketRequestCallback]


class BasicWebSocketRouter(WebSocketRouter):
    """The implementation of a basic Websocket router"""

    def __init__(self) -> None:
        self._routes: List[Route] = []

    def add(self, path: str, callback: WebSocketRequestCallback) -> None:
        self._routes.append((PathDefinition(path), callback))

    def resolve(
            self,
            path: str
    ) -> Tuple[WebSocketRequestCallback, Mapping[str, Any]]:
        for path_definition, handler in self._routes:
            is_match, matches = path_definition.match(path)
            if is_match:
                LOGGER.debug(
                    'Matched "%s"" with %s.',
                    path,
                    path_definition,
                    extra={'path': path}
                )
                return handler, matches

        LOGGER.warning(
            'Failed to find a match for "%s".',
            path,
            extra={'path': path}
        )

        # TODO: Should we have a "route not found" handler?
        raise ValueError(f"Unable to find route for {path}")



"""bareASGI http support"""

from .http_callbacks import (
    HttpRequestCallback,
    HttpMiddlewareCallback,
)
from .http_instance import HttpInstance
from .http_middleware import make_middleware_chain
from .http_request import HttpRequest
from .http_response import HttpResponse, PushResponse
from .http_router import HttpRouter

__all__ = [
    'HttpInstance',
    'HttpRequest',
    'HttpResponse',
    'HttpRouter',
    'HttpRequestCallback',
    'HttpMiddlewareCallback',
    'PushResponse',
    'make_middleware_chain'
]



"""The http request"""

from typing import Any, AsyncIterable, Dict, Mapping

from asgi_typing import HTTPScope
from bareutils import header


class HttpRequest:
    """An HTTP request"""

    def __init__(
            self,
            scope: HTTPScope,
            info: Dict[str, Any],
            context: Dict[str, Any],
            matches: Mapping[str, Any],
            body: AsyncIterable[bytes]
    ) -> None:
        """An HTTP request.

        Args:
            scope (HTTPScope): The ASGI http scope.
            info (Dict[str, Any]): Shared data from the application.
            context (Dict[str, Any]): Private data for this request or chain of
                requests.
            matches (Mapping[str, Any]): Matches made by the router.
            body (AsyncIterable[bytes]): The body.
        """
        self.scope = scope
        self.info = info
        self.context = context
        self.matches = matches
        self.body = body

    @property
    def url(self) -> str:
        """Make the url from the scope.

        Returns:
            str: The url.
        """
        scheme = self.scope['scheme']
        host = header.find(b'host', self.scope['headers'], b'unknown')
        assert host is not None
        path = self.scope['path']
        return f"{scheme}://{host.decode()}{path}"



"""A handler of lifecycle event requests"""

import logging
from typing import Any, Dict, List

from asgi_typing import (
    LifespanScope,
    ASGILifespanReceiveCallable,
    ASGILifespanSendCallable,
    LifespanShutdownCompleteEvent,
    LifespanStartupCompleteEvent,
    LifespanStartupFailedEvent,
    LifespanShutdownFailedEvent
)

from .lifespan_request import LifespanRequest, LifespanRequestHandler

LOGGER = logging.getLogger(__name__)


class LifespanInstance:
    """An instance factor for lifespan event requests"""

    def __init__(
            self,
            scope: LifespanScope,
            startup_handlers: List[LifespanRequestHandler],
            shutdown_handlers: List[LifespanRequestHandler],
            info: Dict[str, Any]
    ) -> None:
        self.scope = scope
        self.startup_handlers = startup_handlers
        self.shutdown_handlers = shutdown_handlers
        self.info = info

    async def process(
            self,
            receive: ASGILifespanReceiveCallable,
            send: ASGILifespanSendCallable
    ) -> None:
        # The lifespan scope exists for the duration of the event loop, and
        # only exits on 'lifespan.shutdown'.
        done = False
        while not done:
            # Fetch the lifespan request
            event = await receive()

            LOGGER.debug('Handling event type "%s".', event['type'])

            if event['type'] == 'lifespan.startup':
                await self._handle_startup_event(send)
            elif event['type'] == 'lifespan.shutdown':
                await self._handle_shutdown_event(send)
                done = True

    async def _handle_startup_event(
            self,
            send: ASGILifespanSendCallable
    ) -> None:
        try:
            await self._call_handlers(
                self.startup_handlers
            )
            startup_complete_event: LifespanStartupCompleteEvent = {
                'type': 'lifespan.startup.complete'
            }
            await send(startup_complete_event)
        except Exception as error:  # pylint: disable=broad-except
            startup_failed_event: LifespanStartupFailedEvent = {
                'type': 'lifespan.startup.failed',
                'message': f'{type(error).__name__}: {error}'
            }
            await send(startup_failed_event)

    async def _handle_shutdown_event(
            self,
            send: ASGILifespanSendCallable
    ) -> None:
        try:
            await self._call_handlers(
                self.shutdown_handlers
            )
            shutdown_complete_event: LifespanShutdownCompleteEvent = {
                'type': 'lifespan.shutdown.complete'
            }
            await send(shutdown_complete_event)
        except Exception as error:  # pylint: disable=broad-except
            shutdown_failed_event: LifespanShutdownFailedEvent = {
                'type': 'lifespan.shutdown.failed',
                'message': f'{type(error).__name__}: {error}'
            }
            await send(shutdown_failed_event)

    async def _call_handlers(
            self,
            handlers: List[LifespanRequestHandler]
    ) -> None:
        for handler in handlers:
            await handler(
                LifespanRequest(
                    self.scope,
                    self.info
                )
            )



"""The http response"""

from __future__ import annotations

import json
from typing import Any, AsyncIterable, Callable, Iterable, List, Optional, Tuple

from bareutils import text_writer

PushResponse = Tuple[str, List[Tuple[bytes, bytes]]]


class HttpResponse:
    """The HTTP response"""

    def __init__(
            self,
            status: int,
            headers: Optional[List[Tuple[bytes, bytes]]] = None,
            body: Optional[AsyncIterable[bytes]] = None,
            pushes: Optional[Iterable[PushResponse]] = None
    ) -> None:
        """The HTTP response.

        Args:
            status (int): The status code.
            headers (Optional[List[Tuple[bytes, bytes]]], optional): The headers
                if any. Mandatory headers will be added if missing. Defaults to
                None.
            body (Optional[AsyncIterable[bytes]], optional): The body, if any.
                Defaults to None.
            pushes (Optional[Iterable[PushResponse]], optional): Server pushes,
                if any. Defaults to None.
        """
        self.status = status
        self.headers = headers
        self.body = body
        self.pushes = pushes

    @classmethod
    def from_text(
            cls,
            text: str,
            *,
            status: int = 200,
            content_type: bytes = b'text/plain',
            headers: Optional[List[Tuple[bytes, bytes]]] = None
    ) -> HttpResponse:
        """Create an HTTP response from a text string.

        Args:
            text (str): A text string.
            headers (Optional[List[Tuple[bytes, bytes]]]): Optional headers. Defaults
                to `None`.
            status (int, optional): An optional status. Defaults to `200`.
            content_type (bytes, optional): An optional content type. Defaults
                to `b'text/plain'`.

        Returns:
            HttpResponse: The built HTTP response.
        """
        return HttpResponse(
            status,
            [(b'content-type', content_type)] + (headers or []),
            text_writer(text)
        )

    @classmethod
    def from_json(
            cls,
            data: Any,
            *,
            status: int = 200,
            content_type: bytes = b'application/json',
            headers: Optional[List[Tuple[bytes, bytes]]] = None,
            dumps: Callable[[Any], str] = json.dumps
    ) -> HttpResponse:
        """Create an HTTP response from data converted to JSON.

        Args:
            data (Any): The data to be converted to JSON.
            headers (Optional[List[Tuple[bytes, bytes]]]): Optional headers. Defaults
                to `None`.
            status (int, optional): An optional status code. Defaults to `200`.
            content_type (bytes, optional): An optional content type. Defaults
                to `b'application/json'`.
            dumps (Callable[[Any], str], optional): An optional function to convert
                the data to a JSON string. Defaults to `json.dumps`.

        Returns:
            HttpResponse: _description_
        """
        return cls.from_text(
            dumps(data),
            status=status,
            content_type=content_type,
            headers=headers
        )



"""The http middleware"""

from functools import partial

from .http_callbacks import HttpRequestCallback, HttpMiddlewareCallback
from .http_request import HttpRequest
from .http_response import HttpResponse


async def _call_handler(
        handler: HttpRequestCallback,
        request: HttpRequest
) -> HttpResponse:
    return await handler(request)


def make_middleware_chain(
        *handlers: HttpMiddlewareCallback,
        handler: HttpRequestCallback
) -> HttpRequestCallback:
    """Create a handler from a chain of middleware.

    Args:
        *handlers (HttpMiddlewareCallback): The middleware handlers.
        handler (HttpRequestCallback): The final response handler.

    Returns:
        HttpRequestCallback: A handler which calls the middleware chain.
    """
    for middleware in reversed(handlers):
        handler = partial(middleware, handler=partial(_call_handler, handler))
    return handler



"""bareASGI websocket support"""

from .websocket import WebSocket
from .websocket_instance import WebSocketInstance
from .websocket_request import WebSocketRequest, WebSocketRequestCallback
from .websocket_router import WebSocketRouter

__all__ = [
    'WebSocket',
    'WebSocketInstance',
    'WebSocketRequest',
    'WebSocketRequestCallback',
    'WebSocketRouter'
]



"""
A segment of a path.
"""

from typing import Optional, Tuple, Any, Callable, Mapping
from datetime import datetime

from ..utils import parse_json_datetime

Converter = Callable[[Any, Optional[str]], Any]


class ParseError(Exception):
    """Exception raised on a parse error"""


def _parse_datetime(value, fmt) -> Optional[datetime]:
    return datetime.strptime(value, fmt) if fmt else parse_json_datetime(value)


CONVERTERS: Mapping[str, Converter] = {
    'str': lambda value, fmt: value,
    'int': lambda value, fmt: int(value),
    'float': lambda value, fmt: float(value),
    'datetime': _parse_datetime,
    'path': lambda value, fmt: value,
}


class PathSegment:
    """A class representing the segment of a path"""

    def __init__(self, segment: str) -> None:
        """Create a path segment
        A path segment can be an absolute name "foo", a variable "{foo}", a
        variable and type "{foo:int}" or a variable, type, and
        format "{foo:datetime:Y-m-dTH:M:S}".

        Valid types are: int, float, str, datetime, path.
        The 'path' type catches all following segments, so '/foo/{rest:path}'
        would match '/foo/bar/grum'.
        """
        self.type: Optional[str] = None
        self.format: Optional[str] = None

        if segment.startswith('{') and segment.endswith('}'):
            self.name, *type_and_format = segment[1:-1].split(':', maxsplit=3)
            if len(type_and_format) == 2:
                self.type, self.format = type_and_format
            elif len(type_and_format) == 1:
                self.type, self.format = type_and_format[0], None
            else:
                self.type, self.format = 'str', None
            if self.type and self.type not in CONVERTERS.keys():
                raise TypeError('Unknown type')
            self.is_variable = True
        elif segment.startswith('{') or segment.endswith('}'):
            raise ParseError("Invalid substitution segment")
        elif '{' in segment or '}' in segment:
            raise ParseError("Literal segment contains invalid characters")
        else:
            self.name = segment
            self.is_variable = False
            self.type = None
            self.format = None

    def match(self, value: str) -> Tuple[bool, Optional[str], Optional[Any]]:
        """Try to match a segment.

        :param value: The path segment to match.
        :return: A tuple of: is_match:bool, variable_name:str, value:any
        """
        if self.is_variable:
            # noinspection PyBroadException
            try:
                converter = CONVERTERS[self.type or 'str']
                value = converter(value, self.format) if self.type else value
                return True, self.name, value
            except ValueError:
                return False, None, None
        else:
            return value == self.name, None, None

    def __str__(self):
        return '<PathSegment: ' \
            f'name="{self.name}"' \
            f', is_variable={self.is_variable}' \
            f', type="{self.type}"' \
            f', format="{self.format}"' \
               '>'

    __repr__ = __str__



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bareasgi',
 'bareasgi.basic_router',
 'bareasgi.http',
 'bareasgi.lifespan',
 'bareasgi.middlewares',
 'bareasgi.websockets']

package_data = \
{'': ['*']}

install_requires = \
['bareutils>=4.0.0,<5.0.0', 'jetblack-asgi-typing>=0.4.0,<0.5.0']

setup_kwargs = {
    'name': 'bareasgi',
    'version': '4.1.0',
    'description': 'A lightweight ASGI framework',
    'long_description': "# bareASGI\n\nA lightweight Python [ASGI](user-guide/asgi) web server framework\n(read the [docs](https://rob-blackbourn.github.io/bareASGI/)).\n\n## Overview\n\nThis is a _bare_ ASGI web server framework. The goal is to provide\na minimal implementation, with other facilities (serving static files, CORS,\nsessions, etc.) being implemented by optional packages.\n\nThe framework is targeted at micro-services which require a light footprint\n(in a container for example), or as a base for larger frameworks.\n\nPython 3.8+ is required.\n\n## Optional Packages\n\n- [bareASGI-cors](https://github.com/rob-blackbourn/bareASGI-cors) for cross origin resource sharing,\n- [bareASGI-static](https://github.com/rob-blackbourn/bareASGI-static) for serving static files,\n- [bareASGI-jinja2](https://github.com/rob-blackbourn/bareASGI-jinja2) for [Jinja2](https://github.com/pallets/jinja) template rendering,\n- [bareASGI-graphql-next](https://github.com/rob-blackbourn/bareASGI-graphql-next) for [GraphQL](https://github.com/graphql-python/graphql-core) and [graphene](https://github.com/graphql-python/graphene),\n- [bareASGI-rest](https://github.com/rob-blackbourn/bareASGI-rest) for REST support,\n- [bareASGI-prometheus](https://github.com/rob-blackbourn/bareASGI-prometheus) for [prometheus](https://prometheus.io/) metrics,\n- [bareASGI-session](https://github.com/rob-blackbourn/bareASGI-session) for sessions.\n\n## Functionality\n\nThe framework provides the basic functionality required for developing a web\napplication, including:\n\n- Http,\n- WebSockets,\n- Routing,\n- Lifecycle,\n- Middleware\n\n## Simple Server\n\nHere is a simple server with a request handler that returns some text.\n\n```python\nimport uvicorn\nfrom bareasgi import Application, HttpRequest, HttpResponse, text_writer\n\nasync def example_handler(request: HttpRequest) -> HttpResponse:\n    return HttpResponse(\n        200,\n        [(b'content-type', b'text/plain')],\n        text_writer('This is not a test')\n    )\n\napp = Application()\napp.http_router.add({'GET'}, '/', example_handler)\n\nuvicorn.run(app, port=9009)\n```\n",
    'author': 'Rob Blackbourn',
    'author_email': 'rob.blackbourn@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/rob-blackbourn/bareasgi',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""The http instance"""

import asyncio
from asyncio import Queue
import logging
from typing import (
    Any,
    AsyncIterable,
    Dict,
    Iterable,
    List,
    Set,
    Tuple,
    cast
)

from asgi_typing import (
    HTTPResponseBodyEvent,
    HTTPScope,
    ASGIHTTPReceiveCallable,
    ASGIHTTPSendCallable,
    HTTPRequestEvent,
    HTTPResponseStartEvent,
    HTTPServerPushEvent
)

from ..utils import NullIter

from .http_callbacks import HttpMiddlewareCallback
from .http_errors import HttpInternalError, HttpDisconnectError
from .http_request import HttpRequest
from .http_response import HttpResponse, PushResponse
from .http_router import HttpRouter
from .http_middleware import make_middleware_chain

LOGGER = logging.getLogger(__name__)


class BodyIterator:
    """Iterate over the body content"""

    def __init__(
            self,
            receive: ASGIHTTPReceiveCallable,
            body: bytes,
            more_body: bool
    ) -> None:
        """Initialise the body iterator

        Args:
            receive (Receive): The receive callable
            body (bytes): The initial body
            more_body (bool): Signifies if there is additional content to come.
        """
        self._receive = receive
        self._queue: Queue = Queue()
        self._queue.put_nowait(body)
        self._more_body = more_body

    def __aiter__(self):
        return self

    async def __anext__(self):
        if not self._queue.empty():
            body = await self._queue.get()
            return body

        if not self._more_body:
            raise StopAsyncIteration

        return await self._read()

    async def _read(self) -> bytes:
        event = await self._receive()
        LOGGER.debug('Received event type "%s".', event['type'])

        if event['type'] == 'http.disconnect':
            raise HttpDisconnectError

        if event['type'] != 'http.request':
            LOGGER.error(
                'Failed to understand event type "%s".', event['type'])
            raise HttpInternalError

        request_event = cast(HTTPRequestEvent, event)
        body = request_event.get('body', b'')
        self._more_body = request_event.get('more_body', False)
        return body

    async def flush(self) -> None:
        """Flush all remaining http.request messages"""
        while self._more_body:
            await self._queue.put(await self._read())


class HttpInstance:
    """An HTTP instance services an HTTP request."""

    def __init__(
            self,
            scope: HTTPScope,
            router: HttpRouter,
            middleware: Iterable[HttpMiddlewareCallback],
            info: Dict[str, Any]
    ) -> None:
        self.scope = scope
        self.info = info

        # Find the route.
        self.handler, self.matches = router.resolve(
            scope['method'],
            scope['path']
        )

        # Assemble any middleware.
        if middleware:
            self.handler = make_middleware_chain(
                *middleware,
                handler=self.handler
            )

    async def process(
            self,
            receive: ASGIHTTPReceiveCallable,
            send: ASGIHTTPSendCallable
    ) -> None:

        LOGGER.debug('Start handling request.')

        try:
            request_event = await self._receive_request(receive)

            response = await self._handle_request(receive, request_event)

            await self._send_response(receive, send, response)

        except asyncio.CancelledError:
            pass

    async def _receive_request(
            self,
            receive: ASGIHTTPReceiveCallable,
    ) -> HTTPRequestEvent:
        event = await receive()
        if event['type'] != 'http.request':
            raise HttpInternalError('Expected http.request')
        return cast(HTTPRequestEvent, event)

    async def _handle_request(
            self,
            receive: ASGIHTTPReceiveCallable,
            request_event: HTTPRequestEvent
    ) -> HttpResponse:
        body = BodyIterator(
            receive,
            request_event.get('body', b''),
            request_event.get('more_body', False)
        )
        request = HttpRequest(
            self.scope,
            self.info,
            {},
            self.matches,
            body
        )

        response = await self.handler(request)

        # Typically the request handler has already processed the request
        # body, but we flush all the "http.request" messages so we can catch
        # the final "http.disconnect".
        await body.flush()

        return response

    async def _send_response(
            self,
            receive: ASGIHTTPReceiveCallable,
            send: ASGIHTTPSendCallable,
            response: HttpResponse
    ) -> None:
        send_task = asyncio.create_task(
            self._send_response_events(send, response)
        )
        receive_task = asyncio.create_task(receive())
        pending: Set[asyncio.Future] = {send_task, receive_task}

        is_connected = True

        while is_connected:

            done, pending = await asyncio.wait(
                pending,
                return_when=asyncio.FIRST_COMPLETED
            )

            if receive_task in done:
                event = receive_task.result()
                LOGGER.debug('Received event type "%s".', event)

                # Cancel pending tasks.
                for task in pending:
                    try:
                        task.cancel()
                        await task
                    except:  # pylint: disable=bare-except
                        pass

                # Check for abnormal disconnection.
                if event['type'] != 'http.disconnect':
                    raise HttpInternalError(
                        f'Unexpected request type "{event["type"]}"'
                    )

                LOGGER.debug('Disconnecting.')

                is_connected = False
            elif send_task in done:
                # Fetch result to trigger possible exceptions
                send_task.result()

        LOGGER.debug('Finish handling request.')

    async def _send_response_events(
            self,
            send: ASGIHTTPSendCallable,
            response: HttpResponse
    ) -> None:
        await self._send_response_start_event(
            send,
            response.status,
            response.headers or []
        )

        if response.pushes is not None and self._is_http_push_supported:
            await self._send_response_push_event(send, response.pushes)

        await self._send_response_body_event(send, response.body or NullIter())

    async def _send_response_start_event(
            self,
            send: ASGIHTTPSendCallable,
            status: int,
            headers: List[Tuple[bytes, bytes]]
    ) -> None:
        response_start_event: HTTPResponseStartEvent = {
            'type': 'http.response.start',
            'status': status,
            'headers': headers
        }

        LOGGER.debug('Sending "http.response.start" with status "%s".', status)
        await send(response_start_event)

    async def _send_response_push_event(
            self,
            send: ASGIHTTPSendCallable,
            pushes: Iterable[PushResponse]
    ) -> None:
        for push_path, push_headers in pushes:
            LOGGER.debug(
                'Sending "http.response.push" for path "%s".',
                push_path
            )
            server_push_event: HTTPServerPushEvent = {
                'type': 'http.response.push',
                'path': push_path,
                'headers': push_headers
            }
            await send(server_push_event)

    async def _send_response_body_event(
            self,
            send: ASGIHTTPSendCallable,
            body: AsyncIterable[bytes]
    ) -> None:
        body_iter = body.__aiter__()
        is_first_time = True
        more_body = True
        buf = b''
        while more_body:
            prev = buf
            try:
                # Try to get more body
                buf = await body_iter.__anext__()
            except StopAsyncIteration:
                # The previous buf was the last
                more_body = False

            if is_first_time and more_body:
                is_first_time = False
                continue

            response_body_event: HTTPResponseBodyEvent = {
                'type': 'http.response.body',
                'body': prev,
                'more_body': more_body
            }
            LOGGER.debug(
                'Sending "http.response.body" with more body "%s".',
                more_body
            )
            await send(response_body_event)

    @property
    def _is_http_push_supported(self) -> bool:
        extensions = self.scope.get('extensions', {})
        return (
            self.scope['http_version'] in ('2', '2.0') and
            extensions is not None and
            'http.response.push' in extensions
        )



"""bareASGI exports"""

from asgi_typing import Scope

from bareutils import (
    text_reader,
    text_writer,
    bytes_reader,
    bytes_writer
)

from .application import Application
from .http import (
    HttpRequest,
    HttpResponse,
    HttpRequestCallback,
    HttpMiddlewareCallback,
    PushResponse,
    make_middleware_chain
)
from .lifespan import LifespanRequest
from .websockets import WebSocket, WebSocketRequest

__all__ = [
    "Scope",

    "text_reader",
    "text_writer",
    "bytes_reader",
    "bytes_writer",

    "Application",

    "HttpRequest",
    "HttpResponse",
    "HttpRequestCallback",
    "HttpMiddlewareCallback",
    "PushResponse",
    "make_middleware_chain",

    "LifespanRequest",

    "WebSocket",
    "WebSocketRequest",
]
